"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
import requests
from programy.utils.logging.ylogger import YLogger


class NewsApiApi:

    def get_news(self, url):
        return requests.get(url)            # pragma: no cover


class NewsArticle:

    def __init__(self):
        self.title = None
        self.description = None
        self.published_at = None
        self.author = None
        self.url = None
        self.url_to_image = None

    def _get_json_attribute(self, data, name, def_value=None):
        if name in data:
            return data[name]
        else:
            YLogger.debug(self, "Attribute [%s] missing from New API Article data", name)
            return def_value

    def parse_json(self, data):
        self.title = self._get_json_attribute(data, "title")
        self.description = self._get_json_attribute(data, "description")
        self.published_at = self._get_json_attribute(data, "publishedAt")
        self.author = self._get_json_attribute(data, "author")
        self.url = self._get_json_attribute(data, "url")
        self.url_to_image = self._get_json_attribute(data, "urlToImage")

    def to_json(self):
        data = {}
        data["title"] = self.title
        data["description"] = self.description
        data["publishedAt"] = self.published_at
        data["author"] = self.author
        data["url"] = self.url
        data["urlToImage"] = self.url_to_image
        return data


# https://newsapi.org/bbc-news-api
class NewsAPI:
    BASE_URL = "https://newsapi.org/v1/articles?source=%s&sortBy=%s&apiKey=%s"

    # Single news feeds
    ABC_NEWS_AU = "abc-news-au"
    AL_JAZEERA_ENGLISH = "al-jazeera-english"
    ARS_TECHNICA = "ars-technica"
    ASSOCIATED_PRESS = "associated-press"
    BBC_NEWS = "bbc-news"
    BBC_SPORT = "bbc-sport"
    BLOOMBERG = "bloomberg"
    BUSINESS_INSIDER = "business-insider"
    BUSINESS_INSIDER_UK = "business-insider-uk"
    BUZZFEED = "buzzfeed"
    CNBC = "cnbc"
    CNN = "cnn"
    DAILY_MAIL = "daily-mail"
    ENGADGET = "engadget"
    ENTERTAINMENT_WEEKLY = "entertainment-weekly"
    ESPN = "espn"
    ESPN_CRIC_INFO = "espn-cric-info"
    FINANCIAL_TIMES = "financial-times"
    FOOTBALL_ITALIA = "football-italia"
    FORTUNE = "fortune"
    FOUR_FOUR_TWO = "four-four-two"
    FOX_SPORTS = "fox-sports"
    GOOGLE_NEWS = "google-news"
    HACKER_NEWS = "hacker-news"
    IGN = "ign"
    INDEPENDENT = "independent"
    MASHABLE = "mashable"
    METRO = "metro"
    MIRROR = "mirror"
    MTV_NEWS = "mtv-news"
    MTV_NEWS_UK = "mtv-news-uk"
    NATIONAL_GEOGRAPHIC = "national-geographic"
    NEW_SCIENTIST = "new-scientist"
    NEWSWEEK = "newsweek"
    NEW_YORK_MAGAZINE = "new-york-magazine"
    NFL_NEWS = "nfl-news"
    POLYGON = "polygon"
    RECODE = "recode"
    REDDIT_R_ALL = "reddit-r-all"
    REUTERS = "reuters"
    TALKSPORT = "talksport"
    TECHCRUNCH = "techcrunch"
    TECHRADAR = "techradar"
    THE_ECONOMIST = "the-economist"
    THE_GUARDIAN_AU = "the-guardian-au"
    THE_GUARDIAN_UK = "the-guardian-uk"
    THE_HUFFINGTON_POST = "the-huffington-post"
    THE_NEW_YORK_TIMES = "the-new-york-times"
    THE_NEXT_WEB = "the-next-web"
    THE_SPORT_BIBLE = "the-sport-bible"
    THE_TELEGRAPH = "the-telegraph"
    THE_VERGE = "the-verge"
    THE_WALL_STREET_JOURNAL = "the-wall-street-journal"
    THE_WASHINGTON_POST = "the-washington-post"
    TIME = "time"
    USA_TODAY = "usa-today"

    # Collections
    BUSINESS = "business"
    ENTERTAINMENT = " entertainment"
    GAMING = "gaming"
    MUSIC = "music"
    SCIENCE_AND_NATURE = "science_and_nature"
    SPORT = "sport"
    TECHNOLOGY = "technology"
    UK_NEWS = "uk_news"
    UK_NEWSPAPERS = "uk_newspapers"

    def __init__(self, license_keys, news_api_api=None):

        if news_api_api is None:
            NewsAPI._news_api_api = NewsApiApi()
        else:
            NewsAPI._news_api_api = news_api_api

        self.function_mapping = {
            NewsAPI.ABC_NEWS_AU: NewsAPI.abc_news_au,
            NewsAPI.AL_JAZEERA_ENGLISH: NewsAPI.al_jazeera_english,
            NewsAPI.ARS_TECHNICA: NewsAPI.ars_technica,
            NewsAPI.ASSOCIATED_PRESS: NewsAPI.associated_press,
            NewsAPI.BBC_NEWS: NewsAPI.bbc_news,
            NewsAPI.BBC_SPORT: NewsAPI.bbc_sport,
            NewsAPI.BLOOMBERG: NewsAPI.bloomberg,
            NewsAPI.BUSINESS_INSIDER: NewsAPI.business_insider,
            NewsAPI.BUSINESS_INSIDER_UK: NewsAPI.business_insider_uk,
            NewsAPI.BUZZFEED: NewsAPI.buzzfeed,
            NewsAPI.CNBC: NewsAPI.cnbc,
            NewsAPI.CNN: NewsAPI.cnn,
            NewsAPI.DAILY_MAIL: NewsAPI.daily_mail,
            NewsAPI.ENGADGET: NewsAPI.engadget,
            NewsAPI.ENTERTAINMENT_WEEKLY: NewsAPI.entertainment_weekly,
            NewsAPI.ESPN: NewsAPI.espn,
            NewsAPI.ESPN_CRIC_INFO: NewsAPI.espn_cric_info,
            NewsAPI.FINANCIAL_TIMES: NewsAPI.financial_times,
            NewsAPI.FOOTBALL_ITALIA: NewsAPI.football_italia,
            NewsAPI.FORTUNE: NewsAPI.fortune,
            NewsAPI.FOUR_FOUR_TWO: NewsAPI.four_four_two,
            NewsAPI.FOX_SPORTS: NewsAPI.fox_sports,
            NewsAPI.GOOGLE_NEWS: NewsAPI.google_news,
            NewsAPI.HACKER_NEWS: NewsAPI.hacker_news,
            NewsAPI.IGN: NewsAPI.ign,
            NewsAPI.INDEPENDENT: NewsAPI.independent,
            NewsAPI.MASHABLE: NewsAPI.mashable,
            NewsAPI.METRO: NewsAPI.metro,
            NewsAPI.MIRROR: NewsAPI.mirror,
            NewsAPI.MTV_NEWS: NewsAPI.mtv_news,
            NewsAPI.MTV_NEWS_UK: NewsAPI.mtv_news_uk,
            NewsAPI.NATIONAL_GEOGRAPHIC: NewsAPI.national_geographic,
            NewsAPI.NEW_SCIENTIST: NewsAPI.new_scientist,
            NewsAPI.NEWSWEEK: NewsAPI.newsweek,
            NewsAPI.NEW_YORK_MAGAZINE: NewsAPI.new_york_magazine,
            NewsAPI.NFL_NEWS: NewsAPI.nfl_news,
            NewsAPI.POLYGON: NewsAPI.polygon,
            NewsAPI.RECODE: NewsAPI.recode,
            NewsAPI.REDDIT_R_ALL: NewsAPI.reddit,
            NewsAPI.REUTERS: NewsAPI.reuters,
            NewsAPI.TALKSPORT: NewsAPI.talksport,
            NewsAPI.TECHCRUNCH: NewsAPI.techcrunch,
            NewsAPI.TECHRADAR: NewsAPI.techradar,
            NewsAPI.THE_ECONOMIST: NewsAPI.the_economist,
            NewsAPI.THE_GUARDIAN_AU: NewsAPI.the_guardian_au,
            NewsAPI.THE_GUARDIAN_UK: NewsAPI.the_guardian_uk,
            NewsAPI.THE_HUFFINGTON_POST: NewsAPI.the_huffington_post,
            NewsAPI.THE_NEW_YORK_TIMES: NewsAPI.the_new_york_times,
            NewsAPI.THE_NEXT_WEB: NewsAPI.the_next_web,
            NewsAPI.THE_SPORT_BIBLE: NewsAPI.the_sport_bible,
            NewsAPI.THE_TELEGRAPH: NewsAPI.the_telegraph,
            NewsAPI.THE_VERGE: NewsAPI.the_verge,
            NewsAPI.THE_WALL_STREET_JOURNAL: NewsAPI.the_wall_street_journal,
            NewsAPI.THE_WASHINGTON_POST: NewsAPI.the_washington_post,
            NewsAPI.TIME: NewsAPI.time,
            NewsAPI.USA_TODAY: NewsAPI.usa_today,
            NewsAPI.BUSINESS: NewsAPI.business,
            NewsAPI.ENTERTAINMENT: NewsAPI.entertainment,
            NewsAPI.GAMING: NewsAPI.gaming,
            NewsAPI.MUSIC: NewsAPI.music,
            NewsAPI.SCIENCE_AND_NATURE: NewsAPI.science_and_nature,
            NewsAPI.SPORT: NewsAPI.sport,
            NewsAPI.TECHNOLOGY: NewsAPI.technology,
            NewsAPI.UK_NEWS: NewsAPI.uk_news,
            NewsAPI.UK_NEWSPAPERS: NewsAPI.uk_newspapers,
        }

        if license_keys is None:
            raise Exception("Missing license keys")

        if license_keys.has_key('NEWSAPI_API_KEY'):
            self.api_key = license_keys.get_key('NEWSAPI_API_KEY')
        else:
            raise Exception("No valid license key NEWSAPI_API_KEY found")

    @staticmethod
    def _format_url(service, api_key, sort_by="top"):
        return NewsAPI.BASE_URL % (service, sort_by, api_key)

    @staticmethod
    def _get_data(url_str, api_key, max_articles, sort, reverse):
        url = NewsAPI._format_url(url_str, api_key)
        return NewsAPI._get_news_feed_articles(url, max_articles, sort, reverse)

    @staticmethod
    def _get_news_feed_articles(url, max_articles, sort, reverse):
        YLogger.debug(None, "News API URL: [%s]", url)
        response = NewsAPI._news_api_api.get_news(url)
        articles = []
        if response.status_code == 200:
            header_splits = response.headers['content-type'].split(";")
            if header_splits[0] == 'application/json':
                json_data = response.json()
                if 'articles' in json_data:
                    for article_data in json_data['articles']:
                        article = NewsArticle()
                        article.parse_json(article_data)
                        articles.append(article)
                        YLogger.debug(None, article.description)

                    if sort is True:
                        YLogger.debug(None, "Sorting articles,, reverse=%s", str(reverse))
                        articles.sort(key=lambda article: article.published_at, reverse=reverse)

                    if max_articles != 0:
                        YLogger.debug(None, "Returning max_articles %d articles", max_articles)
                        articles = articles[:max_articles]
                    else:
                        YLogger.debug(None, "Returning all articles")
                else:
                    YLogger.error(None, "NewAPI payload contains no articles attribute")
            else:
                YLogger.error(None, "NewsAPI request none JSON object")

        else:
            YLogger.error(None, "NewsAPI request returned error code %d", response.status_code)

        return articles

    def get_headlines(self, source, max_articles=0, sort=False, reverse=False):

        if source in self.function_mapping:
            function = self.function_mapping[source]
            return function(self.api_key, max_articles, sort, reverse)
        else:
            YLogger.error(self, "No source available for %s", source)
            return []

    @staticmethod
    def to_program_y_text(articles, break_str=" <br /> "):
        return break_str.join("%s - %s" % (article.title, article.description) for article in articles)

    @staticmethod
    def abc_news_au(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.ABC_NEWS_AU, api_key, max_articles, sort, reverse)

    @staticmethod
    def al_jazeera_english(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.AL_JAZEERA_ENGLISH, api_key, max_articles, sort, reverse)

    @staticmethod
    def ars_technica(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.ARS_TECHNICA, api_key, max_articles, sort, reverse)

    @staticmethod
    def associated_press(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.ASSOCIATED_PRESS, api_key, max_articles, sort, reverse)

    @staticmethod
    def bbc_news(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.BBC_NEWS, api_key, max_articles, sort, reverse)

    @staticmethod
    def bbc_sport(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.BBC_SPORT, api_key, max_articles, sort, reverse)

    @staticmethod
    def bloomberg(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.BLOOMBERG, api_key, max_articles, sort, reverse)

    @staticmethod
    def business_insider(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.BUSINESS_INSIDER, api_key, max_articles, sort, reverse)

    @staticmethod
    def business_insider_uk(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.BUSINESS_INSIDER_UK, api_key, max_articles, sort, reverse)

    @staticmethod
    def buzzfeed(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.BUZZFEED, api_key, max_articles, sort, reverse)

    @staticmethod
    def cnbc(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.CNBC, api_key, max_articles, sort, reverse)

    @staticmethod
    def cnn(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.CNN, api_key, max_articles, sort, reverse)

    @staticmethod
    def daily_mail(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.DAILY_MAIL, api_key, max_articles, sort, reverse)

    @staticmethod
    def engadget(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.ENGADGET, api_key, max_articles, sort, reverse)

    @staticmethod
    def entertainment_weekly(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.ENTERTAINMENT_WEEKLY, api_key, max_articles, sort, reverse)

    @staticmethod
    def espn(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.ESPN, api_key, max_articles, sort, reverse)

    @staticmethod
    def espn_cric_info(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.ESPN_CRIC_INFO, api_key, max_articles, sort, reverse)

    @staticmethod
    def financial_times(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.FINANCIAL_TIMES, api_key, max_articles, sort, reverse)

    @staticmethod
    def football_italia(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.FOOTBALL_ITALIA, api_key, max_articles, sort, reverse)

    @staticmethod
    def fortune(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.FORTUNE, api_key, max_articles, sort, reverse)

    @staticmethod
    def four_four_two(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.FOUR_FOUR_TWO, api_key, max_articles, sort, reverse)

    @staticmethod
    def fox_sports(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.FOX_SPORTS, api_key, max_articles, sort, reverse)

    @staticmethod
    def google_news(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.GOOGLE_NEWS, api_key, max_articles, sort, reverse)

    @staticmethod
    def hacker_news(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.HACKER_NEWS, api_key, max_articles, sort, reverse)

    @staticmethod
    def ign(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.IGN, api_key, max_articles, sort, reverse)

    @staticmethod
    def independent(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.INDEPENDENT, api_key, max_articles, sort, reverse)

    @staticmethod
    def mashable(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.MASHABLE, api_key, max_articles, sort, reverse)

    @staticmethod
    def metro(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.METRO, api_key, max_articles, sort, reverse)

    @staticmethod
    def mirror(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.MIRROR, api_key, max_articles, sort, reverse)

    @staticmethod
    def mtv_news(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.MTV_NEWS, api_key, max_articles, sort, reverse)

    @staticmethod
    def mtv_news_uk(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.MTV_NEWS_UK, api_key, max_articles, sort, reverse)

    @staticmethod
    def national_geographic(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.NATIONAL_GEOGRAPHIC, api_key, max_articles, sort, reverse)

    @staticmethod
    def new_scientist(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.NEW_SCIENTIST, api_key, max_articles, sort, reverse)

    @staticmethod
    def newsweek(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.NEWSWEEK, api_key, max_articles, sort, reverse)

    @staticmethod
    def new_york_magazine(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.NEW_YORK_MAGAZINE, api_key, max_articles, sort, reverse)

    @staticmethod
    def nfl_news(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.NFL_NEWS, api_key, max_articles, sort, reverse)

    @staticmethod
    def polygon(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.POLYGON, api_key, max_articles, sort, reverse)

    @staticmethod
    def recode(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.RECODE, api_key, max_articles, sort, reverse)

    @staticmethod
    def reddit(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.REDDIT_R_ALL, api_key, max_articles, sort, reverse)

    @staticmethod
    def reuters(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.REUTERS, api_key, max_articles, sort, reverse)

    @staticmethod
    def talksport(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.TALKSPORT, api_key, max_articles, sort, reverse)

    @staticmethod
    def techcrunch(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.TECHCRUNCH, api_key, max_articles, sort, reverse)

    @staticmethod
    def techradar(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.TECHRADAR, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_economist(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_ECONOMIST, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_guardian_au(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_GUARDIAN_AU, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_guardian_uk(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_GUARDIAN_UK, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_huffington_post(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_HUFFINGTON_POST, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_new_york_times(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_NEW_YORK_TIMES, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_next_web(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_NEXT_WEB, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_sport_bible(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_SPORT_BIBLE, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_telegraph(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_TELEGRAPH, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_verge(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_VERGE, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_wall_street_journal(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_WALL_STREET_JOURNAL, api_key, max_articles, sort, reverse)

    @staticmethod
    def the_washington_post(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.THE_WASHINGTON_POST, api_key, max_articles, sort, reverse)

    @staticmethod
    def time(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.TIME, api_key, max_articles, sort, reverse)

    @staticmethod
    def usa_today(api_key, max_articles, sort, reverse):
        return NewsAPI._get_data(NewsAPI.USA_TODAY, api_key, max_articles, sort, reverse)

    @staticmethod
    def business(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.bloomberg(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.business_insider(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.business_insider_uk(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.cnbc(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.financial_times(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.fortune(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.the_economist(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.the_wall_street_journal(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def entertainment(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.buzzfeed(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.daily_mail(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.entertainment_weekly(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.mashable(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def gaming(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.ign(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.polygon(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def music(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.mtv_news(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.mtv_news_uk(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def science_and_nature(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.national_geographic(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.new_scientist(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def sport(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.bbc_sport(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.espn(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.espn_cric_info(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.football_italia(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.four_four_two(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.fox_sports(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.talksport(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.the_sport_bible(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def technology(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.ars_technica(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.engadget(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.hacker_news(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.recode(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.techcrunch(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.techradar(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.the_verge(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def uk_news(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.bbc_news(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.uk_newspapers(api_key, max_articles, sort, reverse))
        return articles

    @staticmethod
    def uk_newspapers(api_key, max_articles, sort, reverse):
        articles = []
        articles.extend(NewsAPI.the_guardian_uk(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.mirror(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.the_telegraph(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.daily_mail(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.financial_times(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.independent(api_key, max_articles, sort, reverse))
        articles.extend(NewsAPI.metro(api_key, max_articles, sort, reverse))
        return articles



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.clients.events.console.client import ConsoleBotClient
from programy.utils.console.console import outputLog
from programy.storage.utils.processors import CSVFileReader


class AutoAnswerConsoleBotClient(ConsoleBotClient):

    def __init__(self, defaults_qandas, argument_parser=None):
        ConsoleBotClient.__init__(self, argument_parser)
        self._qandas = []
        self.load_qandas(defaults_qandas)

    def process_line(self, name, line):
        del name
        if len(line) > 0:
            self._qandas.append([line[0], [x.strip(' "') for x in line[1:]]])

    def load_qandas(self, defaults_qandas):
        if self.arguments.args.qandas is not None:
            try:
                reader = CSVFileReader(self.arguments.args.qandas)
                reader.process_lines("qandas", self)
                return

            except Exception as excep:
                outputLog(self, "Using q&a defaults, as unable to load file [%s], reason[%s]" %
                          (self.arguments.args.qandas, str(excep)))

        self._qandas = defaults_qandas[:]

    def add_client_arguments(self, parser=None):
        if parser is not None:
            parser.add_argument('--qandas', dest='qandas',
                                help='list of questions and answers in csv format '
                                     '[question, answer1, answer2, ... answern]')

    def is_right_answer(self, response, answers):
        if len(answers) == 1:
            if answers[0] == '*':
                if len(response) > 0:
                    return True

        return bool(response in answers)

    def process_question_answer(self, client_context):
        for qanda in self._qandas:
            question = qanda[0]
            answers = qanda[1]
            response = self.process_question(client_context, question)
            if self.is_right_answer(response, answers) is True:
                outputLog(self, "RIGHT - Question=[%s], Answer=[%s]" % (question, response))

            else:
                outputLog(self, "WRONG - Question=[%s], Answer=[%s]" % (question, response))

        raise KeyboardInterrupt()


if __name__ == '__main__':
    outputLog(None, "Initiating Auto Answer Console Client...")

    default_quands = [
        ["Hello", ["*"]],
        ["What are you", ["*"]],
        ["Where are you", ["I'm currently in Kinghorn."]]
    ]

    console_app = AutoAnswerConsoleBotClient(default_quands)
    console_app.run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader

from programy.storage.entities.store import Store
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.triggers import TriggersStore
from programy.storage.stores.nosql.mongo.dao.trigger import Trigger


class MongoTriggerStore(MongoStore, TriggersStore):
    TriggerS = 'triggers'

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        TriggersStore.__init__(self)

    def collection_name(self):
        return MongoTriggerStore.TriggerS

    def _get_entity(self, name, classname):
        return Trigger(name, classname)

    def load(self, collector, name=None):
        YLogger.info(self, "Loading %s triggers from Mongo", self.collection_name())
        triggers = self.get_all_triggers()
        for trigger in triggers:
            try:
                collector.add_trigger(trigger['name'], ClassLoader.instantiate_class(trigger['trigger_class'])())

            except Exception as excep:
                YLogger.exception(self, "Failed pre-instantiating Trigger [%s]", excep, trigger['trigger_class'])

    def get_all_triggers(self):
        collection = self.collection()
        return collection.find()

    def _load_triggers_from_file(self, filename, verbose):
        count = 0
        success = 0
        with open(filename, "r", encoding="utf-8") as file:
            for line in file:
                if self.process_config_line(line, verbose) is True:
                    success += 1
                count += 1
        return count, success

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        YLogger.info(self, "Uploading %s to Mongo from file [%s]", self.collection_name(), filename)
        try:
            return self._load_triggers_from_file(filename, verbose)

        except Exception as excep:
            YLogger.exception(self, "Error loading file [%s]", excep, filename)

        return 0, 0

    def process_config_line(self, line, verbose=False):
        line = line.strip()
        if line.startswith('#') is False:
            splits = line.split("=")
            if len(splits) > 1:
                trigger_name = splits[0].strip()
                class_name = splits[1].strip()
                trigger = self._get_entity(trigger_name, class_name)
                if verbose is True:
                    YLogger.debug(self, "Loading trigger [%s] = [%s]", trigger_name, class_name)
                return self.add_document(trigger)

        return False





"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from discord import Client as DiscordClient
from programy.utils.logging.ylogger import YLogger
from programy.clients.events.client import EventBotClient
from programy.clients.events.discord.config import DiscordConfiguration
from programy.utils.console.console import outputLog


class DiscordBotClient(EventBotClient):

    def __init__(self, discord_client: DiscordClient, argument_parser=None):
        self._discord_client = discord_client
        EventBotClient.__init__(self, "Discord", argument_parser)

    def get_client_configuration(self):
        return DiscordConfiguration()

    def get_license_keys(self):
        self._access_token = self.license_keys.get_key("DISCORD_TOKEN")

    def wait_and_answer(self):
        self._discord_client.run(self._access_token)

    def on_ready(self):
        YLogger.info(None, "Discord Client server now running")
        outputLog(self, "Discord Client server now running")
        return True

    def ask_question(self, client_context, question):
        response = client_context.bot.ask_question(client_context, question, responselogger=self)
        return self.renderer.render(client_context, response)

    def get_discord_user(self):
        return self._discord_client.user

    def on_message(self, message):
        if message.author != self.get_discord_user():
            try:
                client_context = self.create_client_context(message.author.id)

                self._questions += 1
                return self.ask_question(client_context, message.content)

            except Exception as e:
                YLogger.exception(None, "Failed to ask question", e)

        YLogger.error(self, "Author is Discord User, not allowed")
        return None


if __name__ == '__main__':

    outputLog(None, "Initiating Discord Client...")

    client = DiscordClient()
    discord = DiscordBotClient(client)

    @client.event
    async def on_ready():
        discord.on_ready()

    @client.event
    async def on_message(message):
        response = discord.on_message(message)
        if response:
            channel = message.channel
            if channel:
                await channel.send(response)


    discord.run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.text.text import TextUtils
from programy.parser.exceptions import ParserException, DuplicateGrammarException
from programy.parser.pattern.nodes.oneormore import PatternOneOrMoreWildCardNode
from programy.parser.pattern.nodes.zeroormore import PatternZeroOrMoreWildCardNode


class PatternGraph:

    def __init__(self, aiml_parser, root_node=None):
        self._aiml_parser = aiml_parser
        self._pattern_factory = aiml_parser.brain.pattern_factory

        self._root_node = None
        self._set_root_node(root_node)

    @property
    def root(self):
        return self._root_node

    @property
    def aiml_parser(self):
        return self._aiml_parser

    @property
    def pattern_factory(self):
        return self._pattern_factory

    def _set_root_node(self, root_node):
        if root_node is None:
            YLogger.debug(self, "Defaulting root to PatternRootNode")
            self._root_node = self._pattern_factory.get_root_node()
        else:
            if root_node.is_root() is False:
                raise ParserException("Root node needs to be of base type PatternRootNode")
            self._root_node = root_node

    def empty(self):
        YLogger.debug(self, "Defaulting root to PatternRootNode")
        self._empty_children(self._root_node)
        self._root_node = self._pattern_factory.get_root_node()

    def _empty_children(self, node):
        for child in node.children:
            self._empty_children(child)
            child.children.clear()

    def node_from_text(self, word, userid="*"):
        if word.startswith("$"):
            node_class = self._pattern_factory.new_node_class('priority')
            return node_class(word[1:], userid)
        elif PatternZeroOrMoreWildCardNode.is_wild_card(word):
            node_class = self._pattern_factory.new_node_class('zeroormore')
            return node_class(word, userid)
        elif PatternOneOrMoreWildCardNode.is_wild_card(word):
            node_class = self._pattern_factory.new_node_class('oneormore')
            return node_class(word, userid)
        node_class = self._pattern_factory.new_node_class('word')
        return node_class(word, userid)

    def node_from_element(self, element, userid="*"):

        node_name = TextUtils.tag_from_text(element.tag)
        if self._pattern_factory.exists(node_name) is False:
            raise ParserException("Unknown node name [%s]" % node_name)

        text = None
        if element.text is not None:
            text = TextUtils.strip_whitespace(element.text)

        node_class_instance = self._pattern_factory.new_node_class(node_name)
        node_instance = node_class_instance(element.attrib, text, userid)

        return node_instance

    def _parse_text(self, pattern_text, current_node, userid="*"):

        stripped = pattern_text.strip()

        words = self._aiml_parser.brain.tokenizer.texts_to_words(stripped)

        for word in words:
            if word != '':  # Blank nodes add no value, ignore them
                word = TextUtils.strip_whitespace(word)

                new_node = self.node_from_text(word, userid=userid)

                current_node = current_node.add_child(new_node)

        return current_node

    def get_text_from_element(self, element):
        text = element.text
        if text is not None:
            text = TextUtils.strip_whitespace(text)
            if text == "":
                return None
            return text
        return None

    def get_tail_from_element(self, element):
        text = element.tail
        if text is not None:
            text = TextUtils.strip_whitespace(text)
            if text == "":
                return None
            return text
        return None

    def add_pattern_to_node(self, pattern_element, userid="*"):
        try:

            head_text = self.get_text_from_element(pattern_element)
            if head_text is not None:
                current_node = self._parse_text(head_text, self._root_node, userid=userid)
            else:
                current_node = self._root_node

            for sub_element in pattern_element:
                new_node = self.node_from_element(sub_element)
                current_node = current_node.add_child(new_node)

                tail_text = self.get_tail_from_element(sub_element)
                if tail_text is not None:
                    current_node = self._parse_text(tail_text, current_node)

            return current_node

        except ParserException as parser_excep:
            parser_excep.xml_element = pattern_element
            raise parser_excep

    def add_topic_to_node(self, topic_element, base_node, userid="*"):
        try:

            current_node = self._pattern_factory.new_node_class('topic')(userid)
            current_node = base_node.add_topic(current_node)

            head_text = self.get_text_from_element(topic_element)
            if head_text is not None:
                current_node = self._parse_text(head_text, current_node)

            added_child = False
            for sub_element in topic_element:
                new_node = self.node_from_element(sub_element)
                current_node = current_node.add_child(new_node)

                tail_text = self.get_tail_from_element(sub_element)
                if tail_text is not None:
                    current_node = self._parse_text(tail_text, current_node)
                added_child = True

            if head_text is None:
                if added_child is False:
                    raise ParserException("Topic node text is empty", xml_element=topic_element)

            return current_node

        except ParserException as parser_excep:
            parser_excep.xml_element = topic_element
            raise parser_excep

    def add_that_to_node(self, that_element, base_node, userid="*"):
        try:

            current_node = self._pattern_factory.new_node_class('that')(userid)
            current_node = base_node.add_that(current_node)

            head_text = self.get_text_from_element(that_element)
            if head_text is not None:
                current_node = self._parse_text(TextUtils.strip_whitespace(head_text), current_node)

            added_child = False
            for sub_element in that_element:
                new_node = self.node_from_element(sub_element)
                current_node = current_node.add_child(new_node)

                tail_text = self.get_tail_from_element(sub_element)
                if tail_text is not None:
                    current_node = self._parse_text(tail_text, current_node)
                added_child = True

            if head_text is None:
                if added_child is False:
                    raise ParserException("That node text is empty", xml_element=that_element)

            return current_node

        except ParserException as parser_excep:
            parser_excep.xml_element = that_element
            raise parser_excep

    def add_template_to_node(self, template_graph_root, current_node, userid="*"):
        template_node = self._pattern_factory.new_node_class('template')(template_graph_root, userid)
        current_node = current_node.add_child(template_node, replace_existing=True)
        return current_node

    def add_pattern_to_graph(self, pattern_element, topic_element, that_element, template_graph_root, learn=False,
                             userid="*"):

        pattern_node = self.add_pattern_to_node(pattern_element, userid=userid)

        topic_node = self.add_topic_to_node(topic_element, pattern_node, userid=userid)

        that_node = self.add_that_to_node(that_element, topic_node, userid=userid)

        if that_node.has_template() is True:
            if learn is False:
                if pattern_element.text is not None:
                    raise DuplicateGrammarException("Dupicate grammar tree found [%s]" % (pattern_element.text.strip()))
                else:
                    raise DuplicateGrammarException("Dupicate grammar tree found for bot/set")
            else:
                if pattern_element.text is not None:
                    YLogger.warning(self, "Duplicate grammar tree found [%s] in learn, replacing existing",
                                    pattern_element.text.strip())
                else:
                    YLogger.warning(self, "Duplicate grammar tree found for bot/set in learn, replacing existing")

                self.add_template_to_node(template_graph_root, that_node)
        else:
            self.add_template_to_node(template_graph_root, that_node)

        return that_node

    def count_words_in_patterns(self):
        counter = [0]
        self._count_words_in_children(self._root_node, counter)
        return counter[0]

    def _count_words_in_children(self, node, counter):
        for child in node.children:
            counter[0] += 1
            self._count_words_in_children(child, counter)

    def dump(self, output_func=YLogger.debug, eol="", verbose=True):
        self.root.dump("", output_func, eol, verbose)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Dict
from programy.utils.logging.ylogger import YLogger
from programy.triggers.config import TriggerConfiguration
from programy.context import ClientContext
from programy.triggers.manager import TriggerManager
from programy.storage.factory import StorageFactory


class LocalTriggerManager(TriggerManager):

    def __init__(self, config: TriggerConfiguration):
        TriggerManager.__init__(self, config)
        self._triggers = {}

    @property
    def triggers(self):
        return self._triggers

    def empty(self):
        for event in self._triggers.keys():
            self._triggers[event].clear()
        self._triggers.clear()

    def get_triggers(self, event: str) -> list:
        if event in self._triggers:
            return self._triggers[event]
        return []

    def has_trigger_event(self, event: str) -> bool:
        return bool(event in self._triggers)

    def add_triggers(self, triggers):
        for event, trigger in triggers.items():
            self.add_trigger(event, trigger)

    def add_trigger(self, event: str, trigger):
        try:
            if event not in self._triggers:
                self._triggers[event] = []
            self._triggers[event].append(trigger)

        except Exception as e:
            YLogger.exception(self, "Failed to add trigger [%s]", e, event)

    def _load_trigger_from_store(self, storage_factory):
        trigger_engine = storage_factory.entity_storage_engine(StorageFactory.TRIGGERS)
        triggers_store = trigger_engine.triggers_store()
        triggers_store.load_all(self)

    def load_triggers(self, storage_factory: StorageFactory):

        assert isinstance(storage_factory, StorageFactory)

        YLogger.debug(self, "Loading Triggers")
        if storage_factory.entity_storage_engine_available(StorageFactory.TRIGGERS) is True:
            try:
                self._load_trigger_from_store(storage_factory)

            except Exception as e:
                YLogger.exception(self, "Failed to load triggers from storage", e)

    def _trigger_trigger(self, trigger, client_context, additional, event):
        YLogger.debug(client_context, "Firing trigger for event [%s]", event)
        trigger.trigger(client_context=client_context, additional=additional)
        return True

    def trigger(self, event: str, client_context: ClientContext = None, additional: Dict[str, str] = None) -> bool:

        if client_context is not None:
            assert isinstance(client_context, ClientContext)

        YLogger.debug(client_context, "Event triggered [%s]", event)

        if event in self._triggers:

            if additional is None:
                additional = {}
            if "event" not in additional:
                additional["event"] = event

            trigger_list = self._triggers[event]
            for trigger in trigger_list:
                try:
                    self._trigger_trigger(trigger, client_context, additional, event)
                except Exception as excep:
                    YLogger.exception(client_context, "Trigger %s failed to fire", excep, event)

            return True

        return False



"""
Copyright(c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files(the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.factory import StorageFactory


class SetCollection:

    def __init__(self):
        self._sets = {}
        self._stores = {}

    @property
    def sets(self):
        return self._sets

    @property
    def stores(self):
        return self._stores

    def storename(self, mapname):
        if mapname in self._stores:
            return self._stores[mapname]
        return None

    def empty(self):
        self._sets.clear()
        self._stores.clear()

    def remove(self, set_name):
        self._sets.pop(set_name, None)
        self._stores.pop(set_name, None)

    def add_set(self, set_name, the_set, set_store):

        # Set names always stored in upper case to handle ambiquity
        set_name = set_name.upper()

        if set_name in self._sets:
            raise Exception("Set %s already exists" % set_name)

        YLogger.debug(self, "Adding set [%s][%s] to set group", set_name, set_store)
        self._sets[set_name] = the_set
        self._stores[set_name] = set_store

    def contains(self, name):
        # Set names always stored in upper case to handle ambiquity
        set_name = name.upper()
        return bool(set_name in self._sets)

    def set(self, name):
        # Set names always stored in upper case to handle ambiquity
        set_name = name.upper()
        if set_name in self._sets:
            return self._sets[set_name]
        return None

    def count_words_in_sets(self):
        count = 0
        for _, aset in self._sets.items():
            for value in aset:
                for variant in value:
                    count += len(variant)
        return count

    def _load_collection(self, storage_engine):
        sets_store = storage_engine.sets_store()
        sets_store.load_all(self)

    def load(self, storage_factory):
        if storage_factory.entity_storage_engine_available(StorageFactory.SETS) is True:
            storage_engine = storage_factory.entity_storage_engine(StorageFactory.SETS)
            try:
                self._load_collection(storage_engine)

            except Exception as e:
                YLogger.exception(self, "Failed to load set from storage", e)

        return len(self._sets)

    def _reload_collection(self, storage_engine, set_name):
        sets_store = storage_engine.sets_store()
        sets_store.reload(self, set_name)

    def reload(self, storage_factory, set_name):
        if storage_factory.entity_storage_engine_available(StorageFactory.SETS) is True:
            storage_engine = storage_factory.entity_storage_engine(StorageFactory.SETS)
            try:
                self._reload_collection(storage_engine, set_name)

            except Exception as e:
                YLogger.exception(self, "Failed to load set from storage", e)

        return len(self._sets)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from programy.utils.logging.ylogger import YLogger
from programy.utils.substitutions.substitues import Substitutions


class BaseConfigurationData(ABC):

    def __init__(self, name):

        assert name is not None

        self._section_name = name
        self._additionals = {}

    def exists(self, name):

        assert name is not None

        return bool(name in self._additionals)

    def value(self, key):

        assert key is not None

        if key in self._additionals:
            return self._additionals[key]

        else:
            YLogger.warning(self, "Configuration key [%s] does not exist", key)

        return None

    @property
    def section_name(self):
        return self._section_name

    @section_name.setter
    def section_name(self, name):
        self._section_name = name

    @property
    def id(self):
        return self._section_name

    def sub_bot_root(self, text, root):

        assert text is not None
        assert root is not None

        return text.replace('$BOT_ROOT', root)

    def additionals_to_add(self):
        return []

    def load_additional_key_values(self, configuration, section, subs: Substitutions = None):
        if section is not None:
            for key in configuration.get_keys(section):
                if key in self.additionals_to_add():
                    value = configuration.get_option(section, key, subs=subs)
                    self._additionals[key] = value

    def _extract_license_key(self, attr, license_keys):
        if attr is not None:
            if "LICENSE:" in attr:
                if license_keys.has_key(attr[8:]):
                    return license_keys.get_key(attr[8:])
                else:
                    return None

        return attr

    def check_for_license_keys(self, license_keys):
        del license_keys
        return

    def config_to_yaml(self, data, config, defaults=True):

        assert data is not None
        assert config is not None

        data[config.id] = {}
        config.to_yaml(data[config.id], defaults)

    def to_yaml(self, data, defaults=True):
        for key, value in self._additionals.items():
            data[key] = value




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import requests
from programy.utils.logging.ylogger import YLogger
from programy.services.openchatbot.openchatbot import OpenChatBot
from programy.config.brain.openchatbots import BrainOpenChatBotsConfiguration


class HTTPRequests:

    def get(self, url):
        return requests.get(url)


class OpenChatBotDomainHandler:
    config_file_name = "openchatbot-configuration"
    wellknown_folder = ".well-known"
    default_domain = ["com"]

    def __init__(self, configuration: BrainOpenChatBotsConfiguration, http_requests=None):
        self._config = configuration

        if http_requests is None:
            self._http_requests = HTTPRequests()
        else:
            self._http_requests = http_requests
        self._cached = {}

    def _create_query_url(self, domain, protocol, ext):
        return "%s://%s.%s/%s/%s" % (protocol, domain, ext,
                                     OpenChatBotDomainHandler.wellknown_folder,
                                     OpenChatBotDomainHandler.config_file_name)

    def _try_connection(self, domain, protocol, ext):
        url = self._create_query_url(domain, protocol, ext)

        YLogger.debug(None, "Trying [%s]", url)
        reply = self._http_requests.get(url)

        if reply.status_code == 200:
            domaincb = OpenChatBot.create(domain, reply.json())
            self.cache_domain(domain, domaincb)
            return domaincb

        return None

    def get_extensions(self):
        return OpenChatBotDomainHandler.default_domain + self._config.domains

    def get_cached(self, domain):
        if domain.upper() in self._cached:
            YLogger.info(None, "Returning cached domain chatbot [%s]", domain)
            return self._cached[domain.upper()]
        return None

    def cache_domain(self, domain, domaincb):
        self._cached[domain.upper()] = domaincb

    def get_endpoint(self, domain):

        domaincb = self.get_cached(domain)
        if domaincb is not None:
            return domaincb

        extensions = self.get_extensions()

        # Either http, https, or both
        for protocol in self._config.protocols:
            # always com, then anything in config.domains, e.g co.uk, org, fr, de etc
            for ext in extensions:
                try:
                    domaincb = self._try_connection(domain, protocol, ext)
                    if domaincb is not None:
                        return domaincb

                except Exception as excep:
                    YLogger.exception_nostack(None, "Failed to get endpoint for domain [%s]", excep, domain)

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.processors import ProcessorStore
from programy.storage.stores.sql.dao.processor import PreProcessor
from programy.storage.stores.sql.dao.processor import PostProcessor
from programy.storage.stores.sql.dao.processor import PostQuestionProcessor
from programy.storage.entities.store import Store
from programy.utils.classes.loader import ClassLoader
from programy.utils.console.console import outputLog


class SQLProcessorsStore(SQLStore, ProcessorStore):

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        ProcessorStore.__init__(self)

    def _get_storage_class(self):
        pass    # pragma: no cover

    def load(self, collector, name=None):
        processors = self.get_all_processors()
        for processor in processors:
            try:
                collector.add_processor(ClassLoader.instantiate_class(processor.classname)())
            except Exception as e:
                YLogger.exception(self, "Failed pre-instantiating Processor [%s]", e, processor.classname)

    def _load_processors_from_file(self, filename, verbose):
        count = 0
        success = 0
        with open(filename, "r", encoding="utf-8") as file:
            for line in file:
                if self._process_config_line(line, verbose) is True:
                    success += 1
                count += 1

        return count, success

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        try:
            count, success = self._load_processors_from_file(filename, verbose)

            self.commit(commit)

            return count, success

        except Exception as error:
            YLogger.exception(self, "Failed to load processors from  [%s]", error, filename)

        return 0, 0

    def _process_config_line(self, line, verbose):
        line = line.strip()
        if line[0] != '#':
            processor = self._get_entity(line)
            self.storage_engine.session.add(processor)
            if verbose is True:
                outputLog(self, line)
            return True
        return False

    def _get_entity(self, classname):
        raise NotImplementedError()  # pragma: no cover

    def get_all_processors(self):
        raise NotImplementedError()  # pragma: no cover


class SQLPreProcessorsStore(SQLProcessorsStore, ProcessorStore):

    def __init__(self, storage_engine):
        SQLProcessorsStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(PreProcessor)

    def empty(self):
       return self._get_all()

    def get_all_processors(self):
        return self._storage_engine.session.query(PreProcessor)

    def _get_entity(self, classname):
        return PreProcessor(classname=classname)


class SQLPostProcessorsStore(SQLProcessorsStore, ProcessorStore):

    def __init__(self, storage_engine):
        SQLProcessorsStore.__init__(self, storage_engine)

    def _get_all(self):
        return self.storage_engine.session.query(PostProcessor)

    def empty(self):
        return self._get_all()

    def get_all_processors(self):
        return self._storage_engine.session.query(PostProcessor)

    def _get_entity(self, classname):
        return PostProcessor(classname=classname)


class SQLPostQuestionProcessorsStore(SQLProcessorsStore, ProcessorStore):

    def __init__(self, storage_engine):
        SQLProcessorsStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(PostQuestionProcessor)

    def empty(self):
        return self._get_all()

    def get_all_processors(self):
        return self._storage_engine.session.query(PostQuestionProcessor)

    def _get_entity(self, classname):
        return PostQuestionProcessor(classname=classname)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from textblob import Word
from textblob.wordnet import VERB
from textblob.wordnet import NOUN
from textblob.wordnet import ADJ
from textblob.wordnet import ADV


class Synsets:

    def __init__(self):
        pass

    @staticmethod
    def _get_synsets(string, pos=None):
        if pos is None:
            word = Word(string)
            synsets = word.synsets
        else:
            word = Word(string)
            synsets = word.get_synsets(pos)

        return synsets

    @staticmethod
    def _get_lemma_and_weight(synset):
        parts = synset.name().split(".")
        return parts[0], int(parts[2])

    def get_similar_words(self, string, pos=None, weight=0):
        synsets = Synsets._get_synsets(string, pos)

        synset_words = []
        for synset in synsets:
            synset_lemma, synset_weight = Synsets._get_lemma_and_weight(synset)
            if synset_weight >= weight:
                if synset_lemma not in synset_words:
                    synset_words.append(synset_lemma)

        return synset_words

    def get_similar_verbs(self, string, weight=0):
        return self.get_similar_words(string, VERB, weight=weight)

    def get_similar_nouns(self, string, weight=0):
        return self.get_similar_words(string, NOUN, weight=weight)

    def get_similar_adjectives(self, string, weight=0):
        return self.get_similar_words(string, ADJ, weight=weight)

    def get_similar_adverbs(self, string, weight=0):
        return self.get_similar_words(string, ADV, weight=weight)

    def get_similarity(self, string1, string2, pos=None):
        syns1 = Synsets._get_synsets(string1, pos)
        syns2 = Synsets._get_synsets(string2, pos)

        similarities = []
        for syn1 in syns1:
            for syn2 in syns2:
                similarity = syn1.path_similarity(syn2)
                if similarity is not None:
                    lemma1, _ = Synsets._get_lemma_and_weight(syn1)
                    lemma2, _ = Synsets._get_lemma_and_weight(syn2)
                    similarities.append((lemma1, lemma2, similarity))

        return similarities



import sys
import csv
class CSVToAIMLGenerator(object):

    def __init__(self):
        self._input = sys.argv[1]
        self._output = sys.argv[2]

    def run(self):
        csv_file = None
        aiml_file = None
        try:
            csv_file = open(self._input, "r+")
            csv_reader = csv.reader(csv_file, delimiter=',', quotechar='"')

            aiml_file = open(self._output, "w+")

            aiml_file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            aiml_file.write('<aiml>\n')

            aiml_file.write('\n')
            for line in csv_reader:
                if line:
                    tab = ""
                    pattern = line[0].strip(' "')
                    topic = line[1].strip(' "')
                    that = line[2].strip(' "')
                    template = line[3].strip(' "')

                    if topic != "*":
                        aiml_file.write('<topic name="%s">\n'%topic)
                        tab = "\t"
                    aiml_file.write('%s<category>\n'%tab)
                    aiml_file.write('%s\t<pattern>%s</pattern>\n'%(tab, pattern))
                    if that != "*":
                        aiml_file.write('%s\t<that>%s</that>\n' % (tab, that))
                    aiml_file.write('%s\t<template>\n'%tab)
                    aiml_file.write('%s\t\t%s\n'%(tab, template))
                    aiml_file.write('%s\t</template>\n'%tab)
                    aiml_file.write('%s</category>\n'%tab)

                    if topic != "*":
                        aiml_file.write('</topic>\n')

                    aiml_file.write('\n')

            aiml_file.write('</aiml>\n')

        except Exception as excep:
            print (excep)
        finally:
            if csv_file is not None:
                csv_file.close ()
            if aiml_file is not None:
                aiml_file.flush ()
                aiml_file.close ()

if __name__ == '__main__':

    def run():
        print("Convertin CSV to AIML...")
        generator = CSVToAIMLGenerator()
        generator.run()

    run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
import urllib.request
from programy.utils.logging.ylogger import YLogger
from programy.utils.geo.latlong import LatLong
from programy.utils.text.text import TextUtils


class GoogleAddressComponent:
    def __init__(self):
        self.long_name = None
        self.short_name = None
        self.types = []

    def parse_json(self, data):
        self.long_name = data['long_name']
        self.short_name = data['short_name']
        self.types = []
        for address_type in data['types']:
            self.types.append(address_type)


class GoogleBounds:
    def __init__(self):
        self.northeast = None
        self.southwest = None

    def parse_json(self, data):
        self.northeast = LatLong(data['northeast']['lat'], data['northeast']['lng'])
        self.southwest = LatLong(data['southwest']['lat'], data['southwest']['lng'])


class GoogleGeometry:
    def __init__(self):
        self.location_type = None
        self.location = None
        self.bounds = None
        self.viewport = None

    def parse_json(self, data):
        self.location_type = data['location_type']
        if 'location' in data:
            self.location = LatLong(data['location']['lat'], data['location']['lng'])
        self.bounds = GoogleBounds()
        self.bounds.parse_json(data['bounds'])
        self.viewport = GoogleBounds()
        self.viewport.parse_json(data['bounds'])


class GoogleGeoLocation:
    def __init__(self):
        self.place_id = None
        self.formatted_address = None
        self.address_components = []
        self.geometry = None
        self.types = []

    def parse_json(self, data):
        self.place_id = data['place_id']
        self.formatted_address = data['formatted_address']

        self.address_components = []
        for component_data in data['address_components']:
            component = GoogleAddressComponent()
            component.parse_json(component_data)
            self.address_components.append(component)

        self.geometry = GoogleGeometry()
        self.geometry.parse_json(data['geometry'])

        self.types = []
        for type_data in data['types']:
            self.types.append(type_data)


class GoogelMapsResult:
    def __init__(self):
        self.locations = []
        self.status = None

    def parse_json(self, data):
        self.status = data['status']
        for result in data['results']:
            location = GoogleGeoLocation()
            location.parse_json(result)
            self.locations.append(location)


class GoogleDistance:
    def __init__(self, origin, destination, country="UK", mode="driving", units="imperial"):
        self._origin = origin
        self._destination = destination
        self._country = country
        self._mode = mode
        self._units = units

        self._distance = None
        self._distance_text = None
        self._duration = None
        self._duration_text = None

    @property
    def distance_text(self):
        return self._distance_text

    def parse_json(self, json_data):
        result = json_data[0]

        if 'elements' not in result:
            raise ValueError("Invalid json data array")

        if 'distance' not in result['elements'][0]:
            raise ValueError("Invalid json data array")

        self._distance = result['elements'][0]['distance']['value']
        self._distance_text = result['elements'][0]['distance']['text']

        if 'duration' not in result['elements'][0]:
            raise ValueError("Invalid json data array")

        self._duration = result['elements'][0]['duration']['value']
        self._duration_text = result['elements'][0]['duration']['text']


class DirectionLegStep:
    def __init__(self):
        self._distance = None
        self._distance_text = None
        self._duration = None
        self._duration_text = None
        self._instructions = None

    @property
    def instructions(self):
        return self._instructions

    def parse_json(self, data):
        self._distance = data['distance']['value']
        self._distance_text = data['distance']['text']
        self._duration = data['duration']['value']
        self._duration_text = data['duration']['text']
        self._instructions = TextUtils.strip_html(data['html_instructions'])


class DirectionLeg:
    def __init__(self):
        self._distance = None
        self._distance_text = None
        self._duration = None
        self._duration_text = None
        self._steps = []

    def parse_json(self, data):
        self._distance = data['distance']['value']
        self._distance_text = data['distance']['text']
        self._duration = data['duration']['value']
        self._duration_text = data['duration']['text']
        for step in data['steps']:
            dirlegstep = DirectionLegStep()
            dirlegstep.parse_json(step)
            self._steps.append(dirlegstep)

    def steps_as_a_string(self):
        return ", ".join([step.instructions for step in self._steps])


class GoogleDirections:
    def __init__(self, origin, destination, country="UK", mode="driving", units="imperial"):
        self._origin = origin
        self._destination = destination
        self._country = country
        self._mode = mode
        self._units = units
        self._legs = []

    def parse_json(self, routes):
        route = routes[0]
        for leg in route['legs']:
            dirleg = DirectionLeg()
            dirleg.parse_json(leg)
            self._legs.append(dirleg)

    def legs_as_a_string(self):
        return ", ".join([leg.steps_as_a_string() for leg in self._legs])


class GoogleMaps:
    DIRECTIONS = "http://maps.googleapis.com/maps/api/directions/json?origin={0}&destination={1}" \
                 "&country={2}&sensor=false&mode={3}"
    DISTANCE = "http://maps.googleapis.com/maps/api/distancematrix/json?origins={0}&destinations={1}" \
               "&country={2}&sensor=false&mode={3}&units={4}"
    GEOCODE = "http://maps.google.com/maps/api/geocode/json?address={0}&sensor=false"

    def _make_http_request(self, url):
        response = urllib.request.urlopen(url)      # pragma: no cover
        content = response.read()                   # pragma: no cover
        decoded = content.decode('utf8')            # pragma: no cover
        return json.loads(decoded)                  # pragma: no cover

    def _get_response_as_json(self, url):
        YLogger.debug(self, "GoogleMaps Request = [%s]", url)
        content = self._make_http_request(url)
        YLogger.debug(self, "GoogleMaps Response = [%s]", content)
        return content

    def _get_latlong_for_location_response(self, location):
        location = TextUtils.urlify(location)
        url = GoogleMaps.GEOCODE.format(location)
        return self._get_response_as_json(url)

    def is_error_response(self, response):
        if 'status' in response:
            if response['status'] == 'OVER_QUERY_LIMIT':
                return True
        return False

    def get_latlong_for_location(self, location):
        YLogger.debug(self, "get_latlong_for_location - calling service")
        response = self._get_latlong_for_location_response(location)

        if self.is_error_response(response):
            raise Exception(response['status'])

        geodata = GoogelMapsResult()
        geodata.parse_json(response)

        location = geodata.locations[0]

        return location.geometry.location

    def _get_distance_between_addresses(self, origin, destination, country, mode, units):
        origin = TextUtils.urlify(origin)
        destination = TextUtils.urlify(destination)
        url = GoogleMaps.DISTANCE.format(origin, destination, country, mode, units)
        return self._get_response_as_json(url)

    def get_distance_between_addresses(self, origin, destination, country="UK", mode="driving", units="imperial"):
        YLogger.debug(self, "get_distance_between_addresses - calling service")
        response = self._get_distance_between_addresses(origin, destination, country, mode, units)

        if self.is_error_response(response):
            raise Exception(response['status'])

        if response['status'] == 'OK':
            YLogger.debug(self, "get_distance_between_addresses - OK")
            distance = GoogleDistance(origin, destination, country, mode, units)
            distance.parse_json(response['rows'])
            return distance
        else:
            YLogger.error(self, "get_distance_between_addresses - [%s]", response['status'])
            return None

    def _get_directions_between_addresses_response(self, origin, destination, country, mode, units):
        origin = TextUtils.urlify(origin)
        destination = TextUtils.urlify(destination)
        url = GoogleMaps.DIRECTIONS.format(origin, destination, country, mode, units)
        return self._get_response_as_json(url)

    def get_directions_between_addresses(self, origin, destination, country="UK", mode="driving", units="imperial"):
        YLogger.debug(self, "get_directions_between_addresses - calling live service")
        response = self._get_directions_between_addresses_response(origin, destination, country, mode, units)

        if self.is_error_response(response):
            raise Exception(response['status'])

        if response['status'] == 'OK':
            YLogger.debug(self, "get_directions_between_addresses - OK")
            directions = GoogleDirections(origin, destination, country, mode, units)
            directions.parse_json(response['routes'])
            return directions
        else:
            YLogger.error(self, "get_directions_between_addresses - %s", response['status'])
            return None



import socket
import sys

MSGLEN = 1024

class MySocket:
    """demonstration class only
      - coded for clarity, not efficiency
    """

    def __init__(self, sock=None):
        if sock is None:
            self.sock = socket.socket(
                            socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = sock

    def connect(self, host, port):
        self.sock.connect((host, port))

    def mysend(self, msg):
        totalsent = 0
        while totalsent < MSGLEN:
            sent = self.sock.send(msg[totalsent:])
            if sent == 0:
                return
            totalsent = totalsent + sent

    def myreceive(self):
        chunks = []
        bytes_recd = 0
        while bytes_recd < MSGLEN:
            chunk = self.sock.recv(min(MSGLEN - bytes_recd, 2048))
            if chunk == b'':
                return b''.join(chunks)
            chunks.append(chunk)
            bytes_recd = bytes_recd + len(chunk)


if __name__ == '__main__':

    host = sys.argv[1]
    port = int(sys.argv[2])
    question = '{"userid": "1234567890", "question": "%s}"}' % sys.argv[3]

    sock = MySocket()

    sock.connect(host, port)

    sock.mysend(bytearray(question, encoding ='utf-8'))

    result = sock.myreceive()
    print(result.decode('utf-8'))



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import argparse

from programy.storage.entities.store import Store
from programy.storage.stores.nosql.mongo.engine import MongoStorageEngine
from programy.storage.stores.nosql.mongo.config import MongoStorageConfiguration
from programy.storage.stores.nosql.mongo.store.categories import MongoCategoryStore
from programy.storage.stores.nosql.mongo.store.maps import MongoMapsStore
from programy.storage.stores.nosql.mongo.store.sets import MongoSetsStore
from programy.storage.stores.nosql.mongo.store.rdfs import MongoRDFsStore
from programy.storage.stores.nosql.mongo.store.properties import MongoPropertyStore
from programy.storage.stores.nosql.mongo.store.properties import MongoDefaultVariablesStore
from programy.storage.stores.nosql.mongo.store.properties import MongoRegexesStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoDenormalStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoNormalStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoGenderStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoPersonStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoPerson2Store
from programy.storage.stores.nosql.mongo.store.spelling import MongoSpellingStore
from programy.storage.stores.nosql.mongo.store.licensekeys import MongoLicenseKeysStore
from programy.storage.stores.nosql.mongo.store.usergroups import MongoUserGroupsStore
from programy.storage.stores.nosql.mongo.store.nodes import MongoTemplateNodeStore
from programy.storage.stores.nosql.mongo.store.nodes import MongoPatternNodeStore
from programy.storage.stores.nosql.mongo.store.processors import MongoPreProcessorStore
from programy.storage.stores.nosql.mongo.store.processors import MongoPostProcessorStore
from programy.storage.stores.nosql.mongo.store.processors import MongoPostQuestionProcessorStore
from programy.storage.stores.nosql.mongo.store.triggers import MongoTriggerStore
from programy.storage.stores.nosql.mongo.store.oobs import MongoOOBStore
from programy.utils.console.console import outputLog


class Uploader:

    @staticmethod
    def upload(storetype, url, database, filename, dirname, subdir, extension, verbose=False):
        config = MongoStorageConfiguration()
        config.url = url
        config.database = database
        config.drop_all_first = False
        engine = MongoStorageEngine(config)
        engine.initialise()
        store = Uploader._get_store(storetype, engine)

        outputLog(None, "Emptying [%s]" % storetype)
        store.empty()

        outputLog(None, "Loading [%s]" % storetype)
        if filename is not None:
            count, success = store.upload_from_file(filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=verbose)
            outputLog(None, "Lines processed: %d" % count)
            outputLog(None, "Entities successful: %d" % success)

        elif dirname is not None:
            count, success = store.upload_from_directory(directory=dirname,
                                                         subdir=subdir,
                                                         extension=extension,
                                                         commit=True,
                                                         verbose=verbose)
            outputLog(None, "Lines processed: %d" % count)
            outputLog(None, "Entities successful: %d" % success)

        else:
            raise Exception("You must specify either --file or --dir")

        return count, success

    @staticmethod
    def _get_store(storetype, engine):
        if storetype == 'categories':
            return MongoCategoryStore(engine)
        if storetype == 'maps':
            return MongoMapsStore(engine)
        if storetype == 'sets':
            return MongoSetsStore(engine)
        if storetype == 'rdfs':
            return MongoRDFsStore(engine)
        if storetype == 'preprocessors':
            return MongoPreProcessorStore(engine)
        if storetype == 'postprocessors':
            return MongoPostProcessorStore(engine)
        if storetype == 'postquestionprocessors':
            return MongoPostQuestionProcessorStore(engine)
        if storetype == 'templatenodes':
            return MongoTemplateNodeStore(engine)
        if storetype == 'patternnodes':
            return MongoPatternNodeStore(engine)
        if storetype == 'properties':
            return MongoPropertyStore(engine)
        if storetype == 'defaults':
            return MongoDefaultVariablesStore(engine)
        if storetype == 'regexes':
            return MongoRegexesStore(engine)
        if storetype == 'denormals':
            return MongoDenormalStore(engine)
        if storetype == 'normals':
            return MongoNormalStore(engine)
        if storetype == 'genders':
            return MongoGenderStore(engine)
        if storetype == 'persons':
            return MongoPersonStore(engine)
        if storetype == 'person2s':
            return MongoPerson2Store(engine)
        if storetype == 'spelling':
            return MongoSpellingStore(engine)
        if storetype == 'licenses':
            return MongoLicenseKeysStore(engine)
        if storetype == 'usergroups':
            return MongoUserGroupsStore(engine)
        if storetype == 'triggers':
            return MongoTriggerStore(engine)
        if storetype == 'oobs':
            return MongoOOBStore(engine)

        raise Exception("Unknone entity storetype [%s]" % storetype)

    @staticmethod
    def create_arguments():
        loader_args = argparse.ArgumentParser(description='Program-Y Set Mongo Loader')

        loader_args.add_argument('-e', '--entity', help="Name of entity to load")
        loader_args.add_argument('-u', '--url', help="Mongo DB connection url")
        loader_args.add_argument('-b', '--db', help="Mongo Database")
        loader_args.add_argument('-f', '--file', help="File to load")
        loader_args.add_argument('-d', '--dir', help="Directory to load files from")
        loader_args.add_argument('-x', '--extension', help="Extension of file to load (dir only)")
        loader_args.add_argument('-s', '--subdir', action='store_true', help="Recurse into all subdirectories")
        loader_args.add_argument('-v', '--verbose', action='store_true', help="Verbose output of loading process")

        return loader_args

    @staticmethod
    def get_args(arguments):
        return arguments.parse_args()       #pragma: no cover

    @staticmethod
    def run():
        arguments = Uploader.create_arguments()
        try:
            args = Uploader.get_args(arguments)
            Uploader.upload(args.entity,
                            args.url,
                            args.db,
                            args.file,
                            args.dir,
                            args.subdir,
                            args.extension,
                            verbose=args.verbose)

        except Exception as excep:
            outputLog(None, "Mongo loader error - %s" % excep)
            arguments.print_help()

        except SystemExit as excep:
            outputLog(None, "Mongo loader error - %s" % excep)
            arguments.print_help()

if __name__ == '__main__':

    Uploader.run()                          #pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.utils.logging.ylogger import YLogger
from programy.parser.pattern.nodes.base import PatternNode
from programy.parser.pattern.equalsmatch import EqualsMatch
from programy.parser.exceptions import ParserException


class PatternRegexNode(PatternNode):

    def __init__(self, attribs, text, userid='*'):
        # @TODO This does not handle upper and lower case
        PatternNode.__init__(self, userid)
        self._pattern_text = None
        self._pattern_template = None
        self._pattern = None
        if 'pattern' in attribs:
            self._pattern_text = attribs['pattern']
        elif 'template' in attribs:
            self._pattern_template = attribs['template']
        elif text:
            self._pattern_text = text
        else:
            raise ParserException("Invalid regex node, neither pattern or template specified as attribute or text")

        if self._pattern_text is not None:
            self._pattern = re.compile(self._pattern_text, re.IGNORECASE)

    @property
    def pattern(self):
        return self._pattern

    @property
    def pattern_text(self):
        return self._pattern_text

    @property
    def pattern_template(self):
        return self._pattern_template

    def is_regex(self):
        return True

    def to_xml(self, client_context, include_user=False):
        string = ""
        if self._pattern_template is not None:
            if include_user is True:
                string += '<regex userid="%s" template="%s">' % (self.userid, self._pattern_template)
            else:
                string += '<regex template="%s">' % self._pattern_template
        else:
            if include_user is True:
                string += '<regex userid="%s" pattern="%s">' % (self.userid, self._pattern_text)
            else:
                string += '<regex pattern="%s">' % self._pattern_text
        string += super(PatternRegexNode, self).to_xml(client_context)
        string += "</regex>\n"
        return string

    def to_string(self, verbose=True):
        if verbose is True:
            if self._pattern_template is not None:
                return "REGEX [%s] [%s] template=[%s]" % (
                    self.userid, self._child_count(verbose), self._pattern_template)
            return "REGEX [%s] [%s] pattern=[%s]" % (self.userid, self._child_count(verbose), self._pattern_text)

        if self._pattern_template is not None:
            return "REGEX template=[%s]" % self._pattern_template
        return "REGEX pattern=[%s]" % self._pattern_text

    def equivalent(self, other):
        if other.is_regex():
            if self.userid == other.userid:
                if self._pattern_template is not None:
                    if other.pattern_template is not None:
                        return bool(self._pattern_template == other.pattern_template)
                else:
                    if other.pattern is not None:
                        return bool(self.pattern == other.pattern)
        return False

    def equals(self, client_context, words, word_no):
        word = words.word(word_no)

        if self.userid != '*':
            if self.userid != client_context.userid:
                return EqualsMatch(False, word_no)

        if self._pattern_template is not None:
            template = client_context.brain.regex_templates.regex(self._pattern_template)
            if template is not None:
                result = template.match(word)
                if result is not None:
                    YLogger.debug(client_context, "Match word [%s] regex", word)
                    return EqualsMatch(True, word_no, word)
                else:
                    YLogger.error(client_context, "No word [%s] matched regex", word)
                    return EqualsMatch(False, word_no)
            else:
                return EqualsMatch(False, word_no)
        else:
            result = self.pattern.match(word)
            if result is not None:
                YLogger.debug(client_context, "Match word [%s] regex", word)
                return EqualsMatch(True, word_no, word)
            else:
                YLogger.error(client_context, "No word [%s] matched regex", word)
                return EqualsMatch(False, word_no)
